-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 1999,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive/db/sql/procs/cl_cachegenericparents.sql
-- SID: 1.7
-- Date Modified: 05/05/03 14:36:46
--
-- Module Description
--   Defines the cl_cacheGenericParents stored procedure.

USE Object
GO

--------------------------------------------------------------------------------
-- cl_cacheGenericParents
PRINT 'Creating procedure cl_cacheGenericParents'
GO

IF EXISTS (SELECT * FROM sysobjects WHERE name = 'cl_cacheGenericParents' AND type = 'P')
  DROP PROCEDURE cl_cacheGenericParents
GO

-------------------------------------------------- {COPYRIGHT-TOP-RM}
-- Licensed Materials - Property of IBM                              
--                                                                   
-- Restricted Materials of IBM                                       
--                                                                   
-- 5724-C51                                                          
--                                                                   
-- (C) Copyright IBM Corp. 1999,2004 All Rights Reserved.            
--                                                                   
-- US Government Users Restricted Rights - Use, duplication or       
-- disclosure restricted by GSA ADP Schedule Contract with           
-- IBM Corp.                                                         
-------------------------------------------------- {COPYRIGHT-END-RM}
-- File: archive/db/sql/procs/cl_cachegenericparents.sql
-- SID: 1.7
-- Date Modified: 05/05/03 14:36:46

--------------------------------------------------------------------------------
-- This procedure loads the Common Listener cache (TBSM database table
-- CL_IDCache) with the current and potential parent object instances of
-- GEM generic resources.
--
-- The loading of the Common Listener cache ensures that duplicate parent
-- resource tree hierarchies will not be created in the TBSM database when new
-- or existing generic resources are introduced to the Common Listener through
-- the Common Listener object discovery process.
--------------------------------------------------------------------------------
CREATE PROCEDURE cl_cacheGenericParents
  @verbose  TINYINT = 0
AS
  DECLARE @tcount INT
  SET @tcount = @@TRANCOUNT
  IF @tcount = 0
    BEGIN TRANSACTION cl_cacheGenericParents
  ELSE
    SAVE TRANSACTION cl_cacheGenericParents
  SET NOCOUNT ON

  DECLARE @rc INT, @cnt INT, @sql NVARCHAR(255), @statusmsg NVARCHAR(1024), @curTime nvarchar(32)

  SET @statusmsg = 'cl_cacheGenericParents: Processing starting...'
  EXEC cl_logStatus 0, 0, @statusmsg, 'OK'


  ------------------------------------------------------------------------------
  ------------------------------------------------------------------------------
  -- Build temp table #cl_genericParentClasses containing all resource classes
  -- that are valid immediate parents of generic resource classes.
  ------------------------------------------------------------------------------
  ------------------------------------------------------------------------------
  CREATE TABLE #cl_genericParentClasses
  (
    genericParentClassCid    ClassID       NOT NULL,
    genericParentClassCname  NVARCHAR(24)  NOT NULL
  )

  -- Sample a generic resource class for this installation
  DECLARE @genericCid ClassID
  SELECT @genericCid = MIN(cid)
    FROM GEMLookupCID
    WHERE comptype = 'generic'
      AND enabled = 1

  IF @genericCid IS NOT NULL
  BEGIN
    -- Get the current list of physical hierarchy classes that are valid immediate parents of generic resources classes for this installation
    INSERT INTO #cl_genericParentClasses
      SELECT oc.cid, oc.cname
        FROM obj_class AS oc WITH (NOLOCK)
        JOIN derived_obj_link AS dol WITH (NOLOCK) ON dol.src_cid = oc.cid
        JOIN link_type AS lt WITH (NOLOCK) ON lt.link_type = dol.link_type
        WHERE dol.dst_cid = @genericCid
          AND (lt.flag_containment = 1 OR lt.flag_parent_child = 1)
          AND oc.cid <> 'LOB'
  END
  ELSE
  BEGIN
    -- No enabled generic classes.  Default to the expected list of physical hierarchy classes that are valid immediate parents of generic resources classes.
    INSERT INTO #cl_genericParentClasses
      SELECT cid, cname
        FROM obj_class WITH (NOLOCK)
        WHERE cid = 'NODE'
          OR cid = 'NTSR'
          OR cid = 'UNIX'
  END


  ------------------------------------------------------------------------------
  ------------------------------------------------------------------------------
  -- Build temp table #cl_genImmediateParents containing the current and
  -- potential immediate parent objects of generic resources.
  ------------------------------------------------------------------------------
  ------------------------------------------------------------------------------
  CREATE TABLE #cl_genericImmediateParents
  (
    genericParentOid     ObjID          NOT NULL,
    genericParentCid     ClassID        NOT NULL,
    genericParentName    ObjName        NOT NULL
  )

  DECLARE @genericParentClassCid ClassID, @genericParentClassCname NVARCHAR(24)

  DECLARE @class_fs INT
  DECLARE class_c CURSOR FOR
    SELECT genericParentClassCid, genericParentClassCname
      FROM #cl_genericParentClasses
  OPEN class_c
  FETCH NEXT FROM class_c INTO @genericParentClassCid, @genericParentClassCname
  SET @class_fs = @@FETCH_STATUS

  -- Determine the current and potential immediate parent objects.
  -- Loop for each parent class in the #cl_genParentClasses table.
  WHILE @class_fs <> -1
  BEGIN
    -- Get all available objects for the parent class
    SET @sql = 'INSERT INTO #cl_genericImmediateParents' +
               ' SELECT c.id, c.cid, c.name' +
               ' FROM ' + @genericParentClassCname + '_C AS c' +
               ' WHERE c.deleted = 0'
    EXEC sp_executesql @sql

    -- Log parent resource statistics
    SELECT @cnt = COUNT(*)
      FROM #cl_genericImmediateParents
      WHERE genericParentCid = @genericParentClassCid

    SET @statusmsg = 'cl_cacheGenericParents: ' + @genericParentClassCname + ' (' + @genericParentClassCid + ') objects found = <' + CAST(@cnt AS NVARCHAR(9)) + '>'
    EXEC cl_logStatus 0, 0, @statusmsg, 'OK'

    FETCH NEXT FROM class_c INTO @genericParentClassCid, @genericParentClassCname
    SET @class_fs = @@FETCH_STATUS

  END  -- End WHILE loop for each parent class in the #cl_genParentClasses table

  DEALLOCATE class_c

  SELECT @cnt = COUNT(*) FROM #cl_genericImmediateParents
  SET @statusmsg = 'cl_cacheGenericParents: Total generic resource immediate parent objects found = <' + CAST(@cnt AS NVARCHAR(9)) + '>'
  EXEC cl_logStatus 0, 0, @statusmsg, 'OK'


  ------------------------------------------------------------------------------
  ------------------------------------------------------------------------------
  -- Build temp tables #cl_generation, to record and process an individual
  -- generic parent object's tree hierarchy, and #cl_tempIDCache to record all
  -- parent objects to be loaded into the Common Listener CL_IDCache table.
  ------------------------------------------------------------------------------
  ------------------------------------------------------------------------------
  CREATE TABLE #cl_generation
  (
    parentLevel       INT            NOT NULL,
    parentOid         ObjID          NOT NULL,
    parentCid         ClassID        NOT NULL,
    parentInstid      NVARCHAR(425)  NOT NULL,
    parentPathInstid  NVARCHAR(425)  NULL
  )

  CREATE TABLE #cl_tempIDCache
  (
    oid              ObjID          NOT NULL,
    cid              ClassID        NOT NULL,
    keyword          NVARCHAR(32)   NOT NULL,
    value            NVARCHAR(512)  NOT NULL,
    instrID          NVARCHAR(255)  NOT NULL,
    duplicateInstid  TINYINT        NOT NULL  DEFAULT(0),
    oldInstid        NVARCHAR(512)  NULL
  )

  DECLARE @genericParentOid ObjID, @genericParentCid ClassID, @genericParentName ObjName
  DECLARE @parentOid ObjID, @parentCid ClassID, @parentInstid NVARCHAR(425)
  DECLARE @src_id ObjID, @src_cid ClassID, @dst_id ObjID, @dst_cid ClassID
  DECLARE @src_instid NVARCHAR(425), @cname NVARCHAR(32)
  DECLARE @totalGenerations INT, @generation INT
  DECLARE @moreParents TINYINT, @moreCacheEntries TINYINT, @orphanedObjects TINYINT, @parentLevel INT
  DECLARE @defaultInstrID NVARCHAR(64), @defaultInstidKey NVARCHAR(16), @defaultALInstrID NVARCHAR(64)
  SET @defaultInstrID = 'TBSM_LOCAL_INSTRUMENTATION: cl_cacheGenericParents'
  SET @defaultInstidKey = 'CL_InstID'
  SET @defaultALInstrID = 'TBSM_LOCAL_INSTRUMENTATION: gemsp_get_parent'
  SET @orphanedObjects = 0

  DECLARE @obj_fs INT
  DECLARE obj_c CURSOR FOR
    SELECT genericParentOid, genericParentCid, genericParentName
      FROM #cl_genericImmediateParents
      ORDER BY genericParentCid ASC, genericParentOid ASC
  OPEN obj_c
  FETCH NEXT FROM obj_c INTO @genericParentOid, @genericParentCid, @genericParentName
  SET @obj_fs = @@FETCH_STATUS

  -- Determine the entire parent hierarchy for each immediate parent object of generic resources.
  -- Loop for each immediate parent object in the #cl_genericImmediateParents table.
  WHILE @obj_fs <> -1
  BEGIN
    -- Save the first generation (level) information.
    INSERT INTO #cl_generation (parentLevel, parentOid, parentCid, parentInstid)
      VALUES (1, @genericParentOid, @genericParentCid, @genericParentName)

    -- Set starting values for a parent hierarchy lookup for the fetched immediate parent object
    SET @dst_id = @genericParentOid
    SET @dst_cid = @genericParentCid
    SET @src_id = 0
    SET @src_cid = ''
    SET @parentLevel = 2
    SET @moreParents = 1

    -- Loop for each parent generation of the fetched immediate parent object
    WHILE @moreParents = 1
    BEGIN
      SELECT @src_id = src_id, @src_cid = src_cid
        FROM link AS l WITH (NOLOCK)
        JOIN link_type AS lt WITH (NOLOCK) ON lt.link_type = l.link_type
        WHERE l.dst_id = @dst_id
          AND l.dst_cid = @dst_cid
          AND l.deleted = 0
          AND (lt.flag_containment = 1 OR lt.flag_parent_child = 1)

      IF ((@src_id > 0) AND ((@src_cid IS NOT NULL) AND (@src_cid <> '')))
      BEGIN
        -- Parent generation found.  Check for root Business Container object.
        IF ((@src_id = 1) AND (@src_cid = 'BUSC'))
        BEGIN
          -- Reached the tree root.  Done finding parents for this generic object.
          SET @moreParents = 0
        END
        ELSE
        BEGIN
          -- Non-root parent object.  Get parent instid from class table.
          SELECT @cname = cname
            FROM obj_class AS o
            WHERE o.cid = @src_cid

          SET @sql = 'SELECT @src_instid = c.name' +
                     ' FROM ' + @cname + '_C AS c' +
                     ' WHERE c.id = ' + CAST(@src_id AS NVARCHAR(9)) +
                     ' AND c.deleted = 0'
          EXEC sp_executesql @sql, N'@src_instid NVARCHAR(425) OUTPUT', @src_instid = @src_instid OUTPUT

          IF ((@src_instid IS NOT NULL) AND (@src_instid <> ''))
          BEGIN
            -- Parent instid found.  Save complete generation (level) information.
            INSERT INTO #cl_generation (parentLevel, parentOid, parentCid, parentInstid)
              VALUES (@parentLevel, @src_id, @src_cid, @src_instid)
          END
          ELSE
          BEGIN
            -- Unexpected missing parent instid, resulting in an orphaned object
            SET @orphanedObjects = 1
            SET @statusmsg = 'cl_cacheGenericParents: Orphaned parent object found: ID = <' + CAST(@genericParentOid AS NVARCHAR(9)) + '>, CID = <' + @genericParentCid + '>, instance ID = <' + @genericParentName + '>.  Invalid instance ID found in parent hierarchy for parent object: ID = <' + CAST(@src_id AS NVARCHAR(9)) + '>, CID = <' + @src_cid + '>.'
            EXEC cl_logStatus 0, 0, @statusmsg, 'WARNING'

            SET @moreParents = 0
            GOTO FetchNextGenericParent
          END
        END
      END
      ELSE
      BEGIN
        -- Unexpected missing parent link, resulting in an orphaned object
        SET @orphanedObjects = 1
        SET @statusmsg = 'cl_cacheGenericParents: Orphaned parent object found: ID = <' + CAST(@genericParentOid AS NVARCHAR(9)) + '>, CID = <' + @genericParentCid + '>, instance ID = <' + @genericParentName + '>.  Invalid parent link found in parent hierarchy for object: ID = <' + CAST(@dst_id AS NVARCHAR(9)) + '>, CID = <' + @dst_cid + '>.'
        EXEC cl_logStatus 0, 0, @statusmsg, 'WARNING'

        SET @moreParents = 0
        GOTO FetchNextGenericParent
      END

      IF @moreParents = 1
      BEGIN
        -- Reset values for next parent lookup
        SET @dst_id = @src_id
        SET @dst_cid = @src_cid
        SET @src_id = 0
        SET @src_cid = ''
        SET @parentLevel = @parentLevel + 1
      END
    END  -- End WHILE loop for each parent generation of the fetched generic object

    -- Create the cache entries for each parent generation of the fetched generic object
    SELECT @totalGenerations = COUNT(*) FROM #cl_generation
    SET @generation = @totalGenerations

    -- Loop for each parent generation in the #cl_generation table, generating a path-based instance id
    WHILE @generation > 0
    BEGIN
      IF @generation = @totalGenerations
      BEGIN
        UPDATE g SET parentPathInstid = (SELECT parentInstid
                                           FROM #cl_generation
                                           WHERE parentLevel = @generation)
          FROM #cl_generation AS g
          WHERE g.parentLevel <= @generation
      END
      ELSE
      BEGIN
        UPDATE g SET parentPathInstid = parentPathInstid + '/' + (SELECT parentInstid
                                                                    FROM #cl_generation
                                                                    WHERE parentLevel = @generation)
          FROM #cl_generation AS g
          WHERE g.parentLevel <= @generation
      END

      SET @generation = @generation - 1
    END  -- End WHILE loop for each parent generation in the #cl_generation table

    -- Insert unique parent objects into the temporary cache
    INSERT INTO #cl_tempIDCache (oid, cid, keyword, value, instrID)
      SELECT g.parentOid, g.parentCid, @defaultInstidKey, g.parentPathInstid, @defaultInstrID
        FROM #cl_generation as g
        LEFT OUTER JOIN #cl_tempIDCache AS tc ON tc.oid = g.parentOid
          AND tc.cid = g.parentCid
          AND tc.value = g.parentPathInstid
        WHERE tc.oid IS NULL

    FetchNextGenericParent: 

    -- Clear generation table for the next object's parent hierarchy
    TRUNCATE TABLE #cl_generation

    FETCH NEXT FROM obj_c INTO @genericParentOid, @genericParentCid, @genericParentName
    SET @obj_fs = @@FETCH_STATUS

  END  -- End WHILE loop for each generic object in the #cl_genericParents table

  DEALLOCATE obj_c

  IF @orphanedObjects = 1
  BEGIN
    SET @statusmsg = 'cl_cacheGenericParents: Orphaned parent objects will not be inserted into the CL cache (CL_IDCache table).'
    EXEC cl_logStatus 0, 0, @statusmsg, 'WARNING'
  END

  SELECT @cnt = COUNT(*) FROM #cl_tempIDCache
  SET @statusmsg = 'cl_cacheGenericParents: Total parent hierarchy objects found = <' + CAST(@cnt AS NVARCHAR(9)) + '>.'
  EXEC cl_logStatus 0, 0, @statusmsg, 'OK'


  ------------------------------------------------------------------------------
  ------------------------------------------------------------------------------
  -- Load the Common Listener cache with current and potential immediate parent
  -- objects of generic resources, not currently in the Common Listener cache.
  --
  -- The Common Listener cache will use the class id (cid) and the generated
  -- path-based instance ids as its primary keys.  It is possible, though
  -- unlikely, to have more than one object of a particular class id with the
  -- same generated path-based instance id, each with a different object id.
  -- The Common Listener relies upon unique instance ids within a particular
  -- class type, so only the first occurrence of an instance id (lowest numeric
  -- object id) will be maintained in the cache.
  --
  -- Also, it is possible to rename parent objects from console which requires
  -- the path-based instance id to be updated in the cache.
  ------------------------------------------------------------------------------
  ------------------------------------------------------------------------------
  DECLARE @oid ObjID, @cid ClassID, @instid NVARCHAR(512)
  DECLARE @duplicateInstids TINYINT, @renamedInstids TINYINT
  SET @duplicateInstids = 0
  SET @renamedInstids = 0

  SELECT @cnt = COUNT(*) FROM #cl_tempIDCache
  IF @cnt > 0
  BEGIN
    -- Flag any duplicate path-based instance ids within a particular class id 
    UPDATE tc1 SET duplicateInstid = 1
      FROM #cl_tempIDCache AS tc1
      WHERE tc1.oid > (SELECT MIN(tc2.oid)
                         FROM #cl_tempIDCache AS tc2
                         WHERE tc1.cid = tc2.cid
                           AND tc1.value = tc2.value)

    SET @cnt = @@ROWCOUNT

    -- Log the parent objects that will not be inserted into the cache due to having a duplicate path-based instance ids within a particular class id
    IF @cnt > 0
    BEGIN
      SET @duplicateInstids = 1

      INSERT INTO CL_Status
        SELECT 0 AS trId,
               0 AS evDataId,
               'WARNING' AS status,
               'cl_cacheGenericParents: Duplicate parent object instance id found: ID = <' + CAST(tc.oid AS NVARCHAR(9)) + '>, CID = <' + tc.cid + '>, instance ID = <' + tc.value + '>.' AS disposition,
               CURRENT_TIMESTAMP AS time
        FROM #cl_tempIDCache AS tc
        WHERE tc.duplicateInstid = 1
    END

    IF @duplicateInstids = 1
    BEGIN
      SET @statusmsg = 'cl_cacheGenericParents: Total parent hierarchy objects with duplicate instance ids not inserted into the CL cache (CL_IDCache table) = <' + CAST(@cnt AS NVARCHAR(9)) + '>.'
      EXEC cl_logStatus 0, 0, @statusmsg, 'WARNING'
    END

    -- Flag any existing parent objects that have been renamed since the last execution of this procedure
    UPDATE tc SET tc.oldInstid = c.value
      FROM #cl_tempIDCache AS tc, CL_IDCache AS c
      WHERE tc.oid = c.oid
        AND tc.cid = c.cid
        AND c.keyword = @defaultInstidKey
        AND (c.instrID = @defaultInstrID OR c.instrID = @defaultALInstrID)
        AND tc.value <> c.value

    SET @cnt = @@ROWCOUNT

    IF @cnt > 0
    BEGIN
      SET @renamedInstids = 1

      -- Update the instance ids of any existing parent objects that have been renamed since the last execution of this procedure
      UPDATE c SET c.value = tc.value
        FROM CL_IDCache AS c, #cl_tempIDCache AS tc
        WHERE c.oid = tc.oid
          AND c.cid = tc.cid
          AND c.keyword = @defaultInstidKey
          AND (c.instrID = @defaultInstrID OR c.instrID = @defaultALInstrID)
          AND tc.oldInstid IS NOT NULL

      IF @verbose <> 0
      BEGIN
        -- Log all the parent objects that have had their instance ids updated
        INSERT INTO CL_Status
          SELECT 0 AS trId,
                 0 AS evDataId,
                 'OK' AS status,
                 'cl_cacheGenericParents: Parent object has been renamed: ID = <' + CAST(tc.oid AS NVARCHAR(9)) + '>, CID = <' + tc.cid + '>, new instance ID = <' + tc.value + '>, old instance ID = <' + tc.oldInstid + '>.' AS disposition,
                 CURRENT_TIMESTAMP AS time
          FROM #cl_tempIDCache AS tc
          WHERE tc.oldInstid IS NOT NULL
      END
      ELSE
      BEGIN
        SET @renamedInstids = 2

        -- Log a sampling of parent objects that have had their instance ids updated
        INSERT INTO CL_Status
          SELECT TOP 20
                 0 AS trId,
                 0 AS evDataId,
                 'OK' AS status,
                 'cl_cacheGenericParents: Parent object has been renamed: ID = <' + CAST(tc.oid AS NVARCHAR(9)) + '>, CID = <' + tc.cid + '>, new instance ID = <' + tc.value + '>, old instance ID = <' + tc.oldInstid + '>.' AS disposition,
                 CURRENT_TIMESTAMP AS time
          FROM #cl_tempIDCache AS tc
          WHERE tc.oldInstid IS NOT NULL
      END
    END

    IF @renamedInstids = 1
    BEGIN
      SET @statusmsg = 'cl_cacheGenericParents: Total parent hierarchy objects renamed and updated in the CL cache (CL_IDCache table) = <' + CAST(@cnt AS NVARCHAR(9)) + '>.'
      EXEC cl_logStatus 0, 0, @statusmsg, 'OK'
    END
    ELSE IF @renamedInstids = 2
    BEGIN
      SET @statusmsg = 'cl_cacheGenericParents: Total parent hierarchy objects renamed and updated in the CL cache (CL_IDCache table) = <' + CAST(@cnt AS NVARCHAR(9)) + '>.  A sampling of up to 20 renamed parent objects have been logged.'
      EXEC cl_logStatus 0, 0, @statusmsg, 'OK'
    END

    -- Finally, insert any new parent objects into the cache
    INSERT INTO CL_IDCache (oid, cid, keyword, value, instrID)
      SELECT tc.oid, tc.cid, tc.keyword, tc.value, tc.instrID
        FROM #cl_tempIDCache AS tc
        LEFT OUTER JOIN CL_IDCache AS c ON c.cid = tc.cid
          AND c.keyword = tc.keyword
          AND c.value = tc.value
        WHERE c.oid IS NULL
          AND tc.duplicateInstid = 0

    SET @cnt = @@ROWCOUNT
  END

  SET @statusmsg = 'cl_cacheGenericParents: Total new parent hierarchy objects inserted into the CL cache (CL_IDCache table) = <' + CAST(@cnt AS NVARCHAR(9)) + '>.'
  EXEC cl_logStatus 0, 0, @statusmsg, 'OK'

  SET @statusmsg = 'cl_cacheGenericParents: Processing complete.'
  EXEC cl_logStatus 0, 0, @statusmsg, 'OK'

  -- Commit transaction
  IF @tcount = 0
    COMMIT TRAN cl_cacheGenericParents

GO
